Unit Profile;

Interface

Uses Crt,Dos;

Type
  TCart = (ct2k,ct4k,ct8k,ct16k,ctSuperChip,ctParkerBros,ctMNetwork,
           ctStarpath,ctCBS,ctTigerVision,ctPitfall2,ctActivision,
           ctMegaBoy,ct32In1,ctCommaVid,ctUnknown);
  TController = (crJoystick,crPaddle,crVTP,crDriving,crKeyboard,crBooster,crUnknown);
  PFileType = ^TFileType;
  TFileType = Record
    Name     : PChar; {TCartName;}
    RealName : PChar; {TRealCartName;}
    Cart     : TCart;
    Cont     : TController;
    Pos200   : Integer;
    Pos204   : Integer;
    Pos215   : Integer;
    Pos250   : Integer;
    Pos305   : Integer;
    Next     : PFileType;
  End;

Const
  MenuKey : Array[TCart] Of Char =
   (#13,#13,#13,#13,'R','P','M',#13,#13,'T','2','A','G','3','V',#13);
  NoPosSet = 999;

Var
  PosSet : Boolean;

Procedure MakeProfile(Path: String);
Function  FindProfile(Name: String;
                      Var Controller: TController;
                      Var Pos200: Integer;
                      Var Pos204: Integer;
                      Var Pos215: Integer;
                      Var Pos250: Integer;
                      Var Pos305: Integer;
                      Var FileRec: PFileType): TCart;
Procedure AddToProfile(Const Name: String;
                       Const RealName: String;
                       Const Cart: TCart;
                       Const Controller: TController;
                       Const Pos200: Integer;
                       Const Pos204: Integer;
                       Const Pos215: Integer;
                       Const Pos250: Integer;
                       Const Pos305: Integer);

Implementation

Uses AtariDBG,VGAGUI,Strings,INI;

Const
  InFileName  = 'PCAE.KNW';
  OutFileName = 'PCAE.PRO';

Type
  PIndex = ^TIndex;
  TIndex = Array[0..255] Of Pointer;

Const
  CartID : Array[TCart] Of String[2] =
   ('2K','4K','8K','16','SC','PB','MN','SP','CB','TV','P2','AC','MB','32','CV','??');
  ContID : Array[TController] Of String[2] =
   ('JY','PD','VT','DR','KB','BG','%%');

Var
  FHead       : PFileType;
  FTail       : PFileType;
  OldExitProc : Pointer;
  FileIndex   : PIndex;

Procedure Separate(Var St,St1: String);
{ ------------------------------------------------------------------ }
{ Takes an input string St, separates such that the left portion is  }
{ returned in St1 and the right portion is returned in St.  If the   }
{ string has no spaces, St1 returns the entire string and St returns }
{ a string containing two spaces.                                    }
{ ------------------------------------------------------------------ }
Var I: Integer;
Begin
  I := Pos(' ',St);
  If I <> 0 Then
  Begin
    St1 := UpperCase(Trim(Copy(St,1,I - 1)));
    St  := Trim(Copy(St,I + 1,Length(St) - I));
  End
  Else
  Begin
    St1 := UpperCase(St);
    St  := '  ';
  End;
End; { Separate }

Procedure SortFileList;
Var
  F1,F2 : PFileType;
  St    : String;
  P     : PChar;
{  St1   : TRealCartName;}
  Cart  : TCart;
  Cont  : TController;
  Pos   : Integer;

Begin
  WriteLn('Sorting file names...');
  F1 := FHead;
  While F1^.Next <> Nil Do
  Begin
    St := StrPas(F1^.Name);
    F2 := F1^.Next;
    While F2 <> Nil Do
    Begin
      If StrPas(F2^.Name) < St Then
      Begin
        P            := F1^.Name;
        F1^.Name     := F2^.Name;
        F2^.Name     := P;

        P            := F1^.RealName;
        F1^.RealName := F2^.RealName;
        F2^.RealName := P;

        Cart         := F1^.Cart;
        F1^.Cart     := F2^.Cart;
        F2^.Cart     := Cart;

        Cont         := F1^.Cont;
        F1^.Cont     := F2^.Cont;
        F2^.Cont     := Cont;

        Pos          := F1^.Pos200;
        F1^.Pos200   := F2^.Pos200;
        F2^.Pos200   := Pos;

        Pos          := F1^.Pos204;
        F1^.Pos204   := F2^.Pos204;
        F2^.Pos204   := Pos;

        Pos          := F1^.Pos215;
        F1^.Pos215   := F2^.Pos215;
        F2^.Pos215   := Pos;

        Pos          := F1^.Pos250;
        F1^.Pos250   := F2^.Pos250;
        F2^.Pos250   := Pos;

        Pos          := F1^.Pos305;
        F1^.Pos305   := F2^.Pos305;
        F2^.Pos305   := Pos;

        St := StrPas(F1^.Name);
      End;
      F2 := F2^.Next;
    End; { While }
    F1 := F1^.Next;
  End;
End; { SortFileList }

Procedure LoadFileList(Path: String);
Var
  S      : SearchRec;
  F      : PFileType;
  Cont   : TController;
  Pos200 : Integer;
  Pos204 : Integer;
  Pos215 : Integer;
  Pos250 : Integer;
  Pos305 : Integer;

Begin
  WriteLn('Loading file names...');
  FindFirst(Path + '*.BIN',Archive,S);
  While DOSError = 0 Do
  Begin
    FindProfile(UpperCase(S.Name),Cont,Pos200,Pos204,Pos215,Pos250,Pos305,F);
    If F = Nil Then
    Begin
      If FHead = Nil Then
      Begin
        New(FHead);
        FTail := FHead;
      End
      Else
      Begin
        New(FTail^.Next);
        FTail := FTail^.Next;
      End;
      FTail^.Name     := StrPCopyNew(UpperCase(S.Name));
      FTail^.RealName := StrPCopyNew('');
      FTail^.Cart     := ctUnknown;
      FTail^.Cont     := crUnknown;
      FTail^.Pos200   := NoPosSet;
      FTail^.Pos204   := NoPosSet;
      FTail^.Pos215   := NoPosSet;
      FTail^.Pos250   := NoPosSet;
      FTail^.Pos305   := NoPosSet;
      FTail^.Next     := Nil;
    End;
    FindNext(S);
  End; { While }
  FindFirst(Path + '*.PAL',Archive,S);
  While DOSError = 0 Do
  Begin
    FindProfile(UpperCase(S.Name),Cont,Pos200,Pos204,Pos215,Pos250,Pos305,F);
    If F = Nil Then
    Begin
      If FHead = Nil Then
      Begin
        New(FHead);
        FTail := FHead;
      End
      Else
      Begin
        New(FTail^.Next);
        FTail := FTail^.Next;
      End;
      FTail^.Name     := StrPCopyNew(UpperCase(S.Name));
      FTail^.RealName := StrPCopyNew('');
      FTail^.Cart     := ctUnknown;
      FTail^.Cont     := crUnknown;
      FTail^.Pos200   := NoPosSet;
      FTail^.Pos204   := NoPosSet;
      FTail^.Pos215   := NoPosSet;
      FTail^.Pos250   := NoPosSet;
      FTail^.Pos305   := NoPosSet;
      FTail^.Next     := Nil;
    End;
    FindNext(S);
  End; { While }
End; { LoadFileList }

Procedure Setup;
Begin
  FHead := Nil;
  FTail := Nil;
End; { Setup }

Procedure Cleanup;
Var FW1,FW2: PFileType;
Begin
  FW1 := FHead;
  While FW1 <> Nil Do
  Begin
    FW2 := FW1^.Next;
    Dispose(FW1);
    FW1 := FW2;
  End; { While }
  FHead := Nil;
  FTail := Nil;
  FillChar(FileIndex^,SizeOf(FileIndex^),#0);
End; { Cleanup }

Procedure OutputFile(Path: String);
Var
  F       : PFileType;
  Loaded  : LongInt;
  Load    : LongInt;
  Cart    : TCart;
  B       : Byte;
  I       : Integer;
  Done    : Boolean;
  St      : String;
  C       : Char;
  Lines   : Integer;
  InFile  : File;
  OutFile : Text;
  Buf     : Array[0..255] Of Byte;

Begin
  C := #0;
  Write('Outputting... ');
  Assign(OutFile,OutFileName);
  ReWrite(OutFile);
  WriteLn(OutFile,'; PCAE cartridge profile list');
  WriteLn(OutFile,'; Valid cart codes: 2K 4K 8K 16 SC PB MN SP CB TV P2 AC MB 32 CV ??');
  WriteLn(OutFile,'; Valid controller codes: JY PD VT DR KB %%');
  WriteLn(OutFile);
  F     := FHead;
  Lines := 0;
  While F <> Nil Do
  Begin
    If F^.Name[1] <> C Then
    Begin
      C := F^.Name[1];
      Write(C,' ');
    End;
    If FileExists(Path + StrPas(F^.Name)) Then
    Begin
      Assign(InFile,Path + StrPas(F^.Name));
      Reset(InFile,1);
      Loaded := FileSize(InFile);
      Cart   := ctUnknown;
      If Loaded < 4096 Then Cart := ct2k Else
       If Loaded = 4096 Then Cart := ct4k Else
        If Loaded = 8192 Then Cart := ct8k Else
         If (Loaded Mod 8448) = 0 Then Cart := ctStarpath Else
          If Loaded = 12288 Then Cart := ctCBS Else
           If Loaded = 16384 Then Cart := ct16k Else
            If Loaded > 16384 Then Cart := ctSuperChip;
      If Loaded > 256 Then Load := 256 Else Load := Loaded;
      BlockRead(InFile,Buf,Load);
      Close(InFile);

      If {(Loaded = 8192) Or} (Loaded = 10495) Then
      Begin
        B    := Buf[0];
        Done := False;
        I := 0;
        While (I < 128) And Not Done Do
        Begin
          If Buf[I] <> B Then Done := True;
          Inc(I);
        End; { While }
        If Not Done Then Cart := ctPitfall2;
      End;
      If Loaded = 16384 Then
      Begin
        B := Buf[0];
        Done := False;
        I := 0;
        While (I < 256) And Not Done Do
        Begin
          If Buf[I] <> B Then Done := True;
          Inc(I);
        End; { While }
        If Not Done Then Cart := ctSuperChip;
      End;
      If F^.Cart = ctUnknown Then F^.Cart := Cart;
    End;
    St := StrPas(F^.Name);
    While Length(St) < 15 Do St := St + ' ';
    If (StrLen(F^.RealName) > 0) And
       (StrPas(F^.RealName) <> StrPas(F^.Name)) Then
     St := St + '"' + StrPas(F^.RealName) + '"';
    While Length(St) < 58 Do St := St + ' ';

    I  := 3 * Integer(F^.Cart);
    While I > 0 Do
    Begin
      St := St + ' ';
      Dec(I);
    End; { While }
    St := St + CartID[F^.Cart] + ' ' + ContID[F^.Cont] + ' ' +
          IntToStr(F^.Pos200) + ' ' +
          IntToStr(F^.Pos204) + ' ' +
          IntToStr(F^.Pos215) + ' ' +
          IntToStr(F^.Pos250) + ' ' +
          IntToStr(F^.Pos305);
    WriteLn(OutFile,St);
    F := F^.Next;
    Inc(Lines);
    If (Lines Mod 10) = 0 Then
     WriteLn(OutFile,'; ---------------------------------------------------');
  End; { While }
  Close(OutFile);
  WriteLn;
End; { OutputFile }

Procedure LoadKnownList;
Var
  St,St1  : String;
  I,J     : Integer;
  F       : PFileType;
  Found   : Boolean;
  Cart    : TCart;
  Cont    : TController;
  OutFile : Text;

Begin
  WriteLn('Loading known cartridge profile list...');
  If FileExists(InFileName) Then
  Begin
    Assign(OutFile,InFileName);
    Reset(OutFile);
    While Not Eof(OutFile) Do
    Begin
      ReadLn(OutFile,St);
      St := Trim(St);

      { Exclude remarks and blank lines }

      If (St <> '') And (St[1] <> ';') Then
      Begin

        { Locate the file name and search for an existing record }

        Separate(St,St1);
        F     := FHead;
        Found := False;
        While (F <> Nil) And Not Found Do
        Begin
          If St1 = StrPas(F^.Name) Then Found := True Else F := F^.Next;
        End; { While }
        If Found Then
        Begin

          { Locate the game's real name }

          I := Pos('"',St);
          If I <> 0 Then
          Begin
            St := Copy(St,I + 1,Length(St) - I);
            I := Pos('"',St);
            If I <> 0 Then
            Begin
              St1 := Trim(Copy(St,1,I - 1));
              St  := Trim(Copy(St,I + 1,Length(St) - I));
              If St1 <> '' Then
              Begin
                StrDispose(FTail^.RealName);
                FTail^.RealName := StrPCopyNew(St1);
              End;
            End;
          End;

          { Locate the first cartridge/controller mnemonic }

          Separate(St,St1);
          For Cart := ct2k To ctUnknown Do
           If CartID[Cart] = St1 Then FTail^.Cart := Cart;
          For Cont := crJoystick To crUnknown Do
           If ContID[Cont] = St1 Then FTail^.Cont := Cont;

          { Locate the second cartridge/controller mnemonic }

          Separate(St,St1);
          For Cart := ct2k To ctUnknown Do
           If CartID[Cart] = St1 Then FTail^.Cart := Cart;
          For Cont := crJoystick To crUnknown Do
           If ContID[Cont] = St1 Then FTail^.Cont := Cont;

          { Locate the 200-scan-line position value }

          Separate(St,St1);
          If St1 <> '' Then
          Begin
            Val(St1,I,J);
            If J = 0 Then FTail^.Pos200 := I;
          End;

          { Locate the 204-scan-line position value }

          Separate(St,St1);
          If St1 <> '' Then
          Begin
            Val(St1,I,J);
            If J = 0 Then FTail^.Pos204 := I;
          End;

          { Locate the 215-scan-line position value }

          Separate(St,St1);
          If St1 <> '' Then
          Begin
            Val(St1,I,J);
            If J = 0 Then FTail^.Pos215 := I;
          End;

          { Locate the 250-scan-line position value }

          Separate(St,St1);
          If St1 <> '' Then
          Begin
            Val(St1,I,J);
            If J = 0 Then FTail^.Pos250 := I;
          End;

          { Locate the 305-scan-line position value }

          Separate(St,St1);
          If St1 <> '' Then
          Begin
            Val(St1,I,J);
            If J = 0 Then FTail^.Pos305 := I;
          End;
        End;
      End;
    End; { While }
    Close(OutFile);
  End;
End; { LoadKnownList }

Procedure IndexFileList;
Var
  FWork : PFileType;
  C     : Char;
  St    : String;

Begin
  FillChar(FileIndex^,SizeOf(FileIndex^),#0);
  FWork := FHead;
  C     := '*';   { Choose a character that cannot be in the filename }
  While FWork <> Nil Do
  Begin
    St := StrPas(FWork^.Name);
    If St <> '' Then
    Begin
      If St[1] <> C Then
      Begin
        C                  := St[1];
        FileIndex^[Ord(C)] := FWork;
      End;
    End;
    FWork := FWork^.Next;
  End; { While }
End; { IndexFileList }

Procedure MakeProfile;
Begin
{  Cleanup;
  Setup;}
  LoadFileList(Path);
  SortFileList;
{  LoadKnownList;}
  IndexFileList;
  OutputFile(Path);
End; { MakeProfile }

Procedure LoadProfile;
Var
  St,St1  : String;
  I,J     : Integer;
  F       : PFileType;
  Found   : Boolean;
  Cart    : TCart;
  Cont    : TController;
  OutFile : Text;

Begin
  If FileExists(OutFileName) Then
  Begin
    WriteLn('Loading cartridge profile...');
    Assign(OutFile,OutFileName);
    Reset(OutFile);
    While Not Eof(OutFile) Do
    Begin
      ReadLn(OutFile,St);
      St := Trim(St);

      { Exclude remarks and blank lines }

      If (St <> '') And (St[1] <> ';') Then
      Begin

        { Locate the file name }

        Separate(St,St1);
        If FHead = Nil Then
        Begin
          New(FHead);
          FTail := FHead;
        End
        Else
        Begin
          New(FTail^.Next);
          FTail := FTail^.Next;
        End;
        FTail^.Name     := StrPCopyNew(UpperCase(St1));
        FTail^.RealName := StrPCopyNew('');
        FTail^.Next     := Nil;
        FTail^.Cart     := ctUnknown;
        FTail^.Cont     := crUnknown;
        FTail^.Pos200   := NoPosSet;
        FTail^.Pos204   := NoPosSet;
        FTail^.Pos215   := NoPosSet;
        FTail^.Pos250   := NoPosSet;
        FTail^.Pos305   := NoPosSet;

        { Locate the game's real name }

        I := Pos('"',St);
        If I <> 0 Then
        Begin
          St := Copy(St,I + 1,Length(St) - I);
          I := Pos('"',St);
          If I <> 0 Then
          Begin
            St1 := Trim(Copy(St,1,I - 1));
            St  := Trim(Copy(St,I + 1,Length(St) - I));
            If St1 <> '' Then
            Begin
              StrDispose(FTail^.RealName);
              FTail^.RealName := StrPCopyNew(St1);
            End;
          End;
        End;

        { Locate the first cartridge/controller mnemonic }

        Separate(St,St1);
        For Cart := ct2k To ctUnknown Do
         If CartID[Cart] = St1 Then FTail^.Cart := Cart;
        For Cont := crJoystick To crUnknown Do
         If ContID[Cont] = St1 Then FTail^.Cont := Cont;

        { Locate the second cartridge/controller mnemonic }

        Separate(St,St1);
        For Cart := ct2k To ctUnknown Do
         If CartID[Cart] = St1 Then FTail^.Cart := Cart;
        For Cont := crJoystick To crUnknown Do
         If ContID[Cont] = St1 Then FTail^.Cont := Cont;

        { Locate the 200-scan-line position value }

        Separate(St,St1);
        If St1 <> '' Then
        Begin
          Val(St1,I,J);
          If J = 0 Then FTail^.Pos200 := I;
        End;

        { Locate the 204-scan-line position value }

        Separate(St,St1);
        If St1 <> '' Then
        Begin
          Val(St1,I,J);
          If J = 0 Then FTail^.Pos204 := I;
        End;

        { Locate the 215-scan-line position value }

        Separate(St,St1);
        If St1 <> '' Then
        Begin
          Val(St1,I,J);
          If J = 0 Then FTail^.Pos215 := I;
        End;

        { Locate the 250-scan-line position value }

        Separate(St,St1);
        If St1 <> '' Then
        Begin
          Val(St1,I,J);
          If J = 0 Then FTail^.Pos250 := I;
        End;

        { Locate the 305-scan-line position value }

        Separate(St,St1);
        If St1 <> '' Then
        Begin
          Val(St1,I,J);
          If J = 0 Then FTail^.Pos305 := I;
        End;
      End;
    End; { While }
    Close(OutFile);
  End;
End; { LoadProfile }

{$F+}
Procedure KillProfile;
Var Path: String;
Begin
  If ParamCount > 0 Then
  Begin
    Path := ParamStr(1);
    If (Path <> '') And Not (Path[Length(Path)] In [':','\']) Then Path := Path + '\';
  End
  Else Path := '';
  If PosSet Then OutputFile(Path);
  Cleanup;
  Dispose(FileIndex);
  ExitProc := OldExitProc;
End; { KillProfile }
{$F-}

Function FindProfile(Name: String;
                     Var Controller: TController;
                     Var Pos200: Integer;
                     Var Pos204: Integer;
                     Var Pos215: Integer;
                     Var Pos250: Integer;
                     Var Pos305: Integer;
                     Var FileRec: PFileType): TCart;
Var
  FWork : PFileType;
  Found : Boolean;
  St    : String;

Begin
  Name  := UpperCase(Name);
  Found := False;
  If Name <> '' Then
  Begin
    FWork := FileIndex^[Ord(Name[1])];
    If FWork <> Nil Then St := StrPas(FWork^.Name);
    While (FWork <> Nil) And (St[1] = Name[1]) And Not Found Do
    Begin
      If Name = St Then Found := True Else
      Begin
        FWork := FWork^.Next;
        If FWork <> Nil Then St := StrPas(FWork^.Name);
      End;
    End; { While }
  End;
  If Found Then
  Begin
    FindProfile := FWork^.Cart;
    Controller  := FWork^.Cont;
    Pos200      := FWork^.Pos200;
    Pos204      := FWork^.Pos204;
    Pos215      := FWork^.Pos215;
    Pos250      := FWork^.Pos250;
    Pos305      := FWork^.Pos305;
    FileRec     := FWork;
  End
  Else
  Begin
    FindProfile := ctUnknown;
    Controller  := crUnknown;
    Pos200      := NoPosSet;
    Pos204      := NoPosSet;
    Pos215      := NoPosSet;
    Pos250      := NoPosSet;
    Pos305      := NoPosSet;
    FileRec     := Nil;
  End;
End; { FindProfile }

Procedure AddToProfile(Const Name: String;
                       Const RealName: String;
                       Const Cart: TCart;
                       Const Controller: TController;
                       Const Pos200: Integer;
                       Const Pos204: Integer;
                       Const Pos215: Integer;
                       Const Pos250: Integer;
                       Const Pos305: Integer);
Begin
  If FHead = Nil Then
  Begin
    New(FHead);
    FTail := FHead;
  End
  Else
  Begin
    New(FTail^.Next);
    FTail := FTail^.Next;
  End;
  FTail^.Name     := StrPCopyNew(Name);
  FTail^.RealName := StrPCopyNew(RealName);
  FTail^.Cart     := Cart;
  FTail^.Cont     := Controller;
  FTail^.Pos200   := Pos200;
  FTail^.Pos204   := Pos204;
  FTail^.Pos215   := Pos215;
  FTail^.Pos250   := Pos250;
  FTail^.Pos305   := Pos305;
  FTail^.Next     := Nil;
  SortFileList;
End; { AddToProfile }

Procedure LoadOldProfile;
Var
  St,St1  : String;
  I,J     : Integer;
  F       : PFileType;
  Found   : Boolean;
  Cart    : TCart;
  Cont    : TController;
  OutFile : Text;
  Pos200  : Integer;
  Pos204  : Integer;
  Pos215  : Integer;
  Pos250  : Integer;
  Pos305  : Integer;

Begin
  If FileExists(OutFileName) Then
  Begin
    WriteLn('Loading cartridge profile...');
    Assign(OutFile,'PCAE0.PRO');
    Reset(OutFile);
    While Not Eof(OutFile) Do
    Begin
      ReadLn(OutFile,St);
      St := Trim(St);

      { Exclude remarks and blank lines }

      If (St <> '') And (St[1] <> ';') Then
      Begin

        { Locate the file name }

        Separate(St,St1);
        FindProfile(UpperCase(St1),Cont,Pos200,Pos204,Pos215,Pos250,Pos305,F);
        If F <> Nil Then
        Begin

          { Locate the game's real name }

          I := Pos('"',St);
          If I <> 0 Then
          Begin
            St := Copy(St,I + 1,Length(St) - I);
            I := Pos('"',St);
            If I <> 0 Then
            Begin
              St1 := Trim(Copy(St,1,I - 1));
              St  := Trim(Copy(St,I + 1,Length(St) - I));
              If St1 <> '' Then
              Begin
                StrDispose(F^.RealName);
                F^.RealName := StrPCopyNew(St1);
              End;
            End;
          End;

          { Locate the first cartridge/controller mnemonic }

          Separate(St,St1);
          For Cart := ct2k To ctUnknown Do
           If CartID[Cart] = St1 Then F^.Cart := Cart;
          For Cont := crJoystick To crUnknown Do
           If ContID[Cont] = St1 Then F^.Cont := Cont;

          { Locate the second cartridge/controller mnemonic }

          Separate(St,St1);
          For Cart := ct2k To ctUnknown Do
           If CartID[Cart] = St1 Then F^.Cart := Cart;
          For Cont := crJoystick To crUnknown Do
           If ContID[Cont] = St1 Then F^.Cont := Cont;
        End;
      End;
    End; { While }
    Close(OutFile);
  End;
End; { LoadOldfile }

Begin
  FHead := Nil;
  FTail := Nil;
  New(FileIndex);
  LoadProfile;
{  LoadOldProfile;}
  IndexFileList;
  OldExitProc := ExitProc;
  ExitProc    := @KillProfile;
  PosSet      := False;
End.

// ----------------------------------------------------------------------
// PCAE - PC Atari Emulator - Atari 2600 Emulator for MS-DOS
// Copyright (C) 2000 John Dullea
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
// ----------------------------------------------------------------------

